#line 1 "./src/mta/mta.C"
 /* --------------------------------------------------------------------------
 * Copyright 1992-1994 by Forschungszentrum Informatik (FZI)
 * All rights reserved.
 *
 * You can use and distribute this software under the terms of the license
 * you should have received along with this program.
 * For a copy of the license or for additional information about this software,
 * write to Xcc Software, Durlacher Allee 53, D-76131 Karlsruhe, Germany;
 * Email: obst@xcc-ka.de.
 * --------------------------------------------------------------------------
 */
/* OBST LIBRARY MODULE */
// **************************************************************************
// Module mta
// **************************************************************************
// implements methods of all classes declared in schema "mta" 
// **************************************************************************

#include "obst_progstd.h"
#include "obst.h"
#include "smg.h"
#include "mta_err.h"
#include "trc_mta.h"
#include "mta.h"

#ifdef OBST_HAVE_PROT
#include "prot_use.h"
#endif

// dir_obst.h must be included, since scp "optimizes" the call to
// sos_Object_Directory::root to _sos_Object_Directory::root.
#include "dir_obst.h"
#include "cci_use.h"

#include "mta_obst.h"


// -------------------------------------------------------------------------
// sos_Import
// -------------------------------------------------------------------------
// *************************************************************************
sos_Type_descr _sos_Import::lookup_type (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_String)	name,
					sos_Import_mode imp_mode)
// *************************************************************************
const{  T_PROC ("sos_Import::lookup_type")
   TT (mta_H, T_ENTER);

   sos_Type_descr result = get_module(_OBSThis).get_type_table() [name];
   
   if VALID (result) 
      if (imp_mode > get_mode(_OBSThis))
	 set_mode(_OBSThis,imp_mode);

   TT (mta_H, T_LEAVE);
   return result;
}


// -------------------------------------------------------------------------
// sos_Schema_module
// -------------------------------------------------------------------------
// *************************************************************************
sos_Schema_module_Directory _sos_Schema_module::schema_dir ()
// *************************************************************************
{  T_PROC ("sos_Schema_module::schema_dir")
   TT (mta_H, T_ENTER);

   static sos_String name = _sos_String::create (TEMP_CONTAINER, "sos_schemas");
   static sos_Schema_module_Directory dir =
                sos_Schema_module_Directory::make (NO_OBJECT);

   if (INVALID(dir))
   {
      sos_Container_status root_status = ROOT_CONTAINER.status();

      sos_Bool read_open = sos_Bool((root_status == WRITEABLE) OR
                                    (root_status == READABLE));
      if (NOT read_open)
         ROOT_CONTAINER.open (READING, WAITING);

      sos_Directory root_dir = _sos_Object_Directory::root();
      dir = sos_Schema_module_Directory::make (root_dir[name]);
      if (NOT read_open)
          ROOT_CONTAINER.close ();

      if INVALID(dir)
      {  sos_Container dir_cnt = sos_Container::create();
         dir = sos_Schema_module_Directory::create (dir_cnt, name);
         sos_Bool write_open = sos_Bool(root_status == WRITEABLE);
         if (NOT write_open)
            ROOT_CONTAINER.open (WRITING, WAITING);
         dir_cnt.close();
         root_dir.insert (name, dir);
         if (NOT write_open)
         {
            ROOT_CONTAINER.close ();
            // dir.container().close();              geaendert 10.08.93 (bs)
            // dir.container().open (WRITING, WAITING);
         }
      }
   }
   TT (mta_H, T_LEAVE);
   return dir;
}

// *************************************************************************
sos_Schema_module _sos_Schema_module::lookup (OBST_PARDECL(sos_String) name)
// *************************************************************************
{  T_PROC ("sos_Schema_module::lookup")
   TT (mta_H, T_ENTER);

   static sos_Schema_module_Directory sd = _sos_Schema_module::schema_dir ();

   // Opening is done implicit if necessary, close must not be done
   // Close is performed by mta_commit
   sos_Schema_module sm = sd [name];

   TT (mta_H, T_LEAVE);
   return sm;
}

// *************************************************************************
sos_Schema_module _sos_Schema_module::retrieve (sos_Container schema_ct)
// *************************************************************************
{  T_PROC ("sos_Schema_module::retrieve")
   TT (mta_H, T_ENTER);

   sos_Schema_module sm = sos_Schema_module::make (schema_ct.root_object());

   TT (mta_H, T_LEAVE);
   return sm;
}

// *************************************************************************
sos_Type_descr _sos_Schema_module::lookup_type (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_String) name, 
					       sos_Import_mode imp_mode)
// *************************************************************************

// search for a type declaration with 'name' in the type table
// of the schema module. If no declaration is found search for 'name'
// in the type tables of the imported modules. If the search fails
// return NO_OBJECT. If found in an import, the mode of the import is
// raised to the demanded level imp_mode, if necessary.

const{  T_PROC ("sos_Schema_module::lookup_type")
   TT (mta_H, T_ENTER);

   static char 		     mode_code[2] = { '\0', '\0' };
   static sos_String         key	  = sos_String::create (TEMP_CONTAINER);
   static sos_Object_sos_Object_Mapping_Hash tmp_dir
     = sos_Object_sos_Object_Mapping_Hash::create (TEMP_CONTAINER,
						   FALSE, // list_cursor
						   TRUE,  // key_based_on_equal
						   FALSE);// info_b.o.e
   mode_code[0] = 'A' + (int)imp_mode;

   smg_String _key = smg_String(mode_code) + name;

   sos_Type_descr result;

   if (_key.length() == 1)
   {  tmp_dir.clear();
      result = sos_Type_descr::make (NO_OBJECT);
   }
   else
   {  char* keystr = _key.make_Cstring (SMG_BORROW);
      key.assign_Cstring(keystr);

      result = sos_Type_descr::make(tmp_dir[(sos_Object)key]);
      if INVALID(result)
      {  result = get_type_table(_OBSThis)[name];

	 if INVALID(result) // not found in current schema
	 {
	    sos_Import_List imp_list = get_imports(_OBSThis);
	    agg_iterate (imp_list, sos_Import imp)
	    {  result = imp.lookup_type (name, imp_mode);
	       if VALID(result)
		  break;
	    }
	    agg_iterate_end (imp_list, imp);
	 }

	 if VALID (result) // enter found type into speedup-mapping
	    for (sos_Import_mode im=sos_IMP_NONE; 
		 im < imp_mode; 
		 im = sos_Import_mode (sos_Int(im) +1 ))
	    {  *keystr = 'A' + (int)im;

	       tmp_dir.insert (_sos_String::create(TEMP_CONTAINER, keystr),
			       result);
	    }
      }
   }
   TT (mta_H, T_LEAVE);
   return result;
}


// *************************************************************************
void _sos_Schema_module::install (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Schema_module::install")
   TT (mta_H, T_ENTER);

   sos_Schema_module_Directory sd = _sos_Schema_module::schema_dir ();

   sd.container().open (WRITING, WAITING);
   sos_String str = get_name (_OBSThis);
   if (sd.is_key (str))
   {  // not yet implemented: info that schemas must be recompiled

      sos_Container ct = sd[str].container();
      sd.remove (str);
      ct.open (WRITING, WAITING);
      ct.destroy();
      ct.close();
   }
   sd.insert (str, sos_Schema_module::make(_OBSThis,this));
   sd.container().close();

   TT (mta_H, T_LEAVE);
}

// -------------------------------------------------------------------------
// sos_Identifier
// -------------------------------------------------------------------------
// *************************************************************************
sos_Bool _sos_Identifier::total_equal (OBST_PARDECL(sos_Identifier) x,OBST_PARDECL( sos_Object) o,
				      sos_Eq_kind)
// *************************************************************************
{  T_PROC ("sos_Identifier::total_equal")
   TT (mta_M, T_ENTER);

   sos_Bool result = ( x.get_id().equal (sos_Identifier::make (o).get_id()) );

   TT (mta_M, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Comp_descr
// -------------------------------------------------------------------------
// *************************************************************************
sos_Method _sos_Comp_descr::comp_method (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Class_type) ct,
				        sos_Bool       set,
					sos_Bool       redef /* = TRUE */)
// *************************************************************************
const{  T_PROC("sos_Comp_descr::comp_method")
   TT(mta_H, T_ENTER);

   sos_Method result = set ? get_set_method(_OBSThis)
			   : get_get_method(_OBSThis);
   if (redef)
      result = ct.get_methods().lookup (result); 

   TT(mta_H, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Method
// -------------------------------------------------------------------------
// *************************************************************************
sos_Bool _sos_Method::overloads (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Method) m)
// *************************************************************************
// precondition: self.name == m.name

const{  T_PROC ("sos_Method::overloads")
   TT (mta_H, T_ENTER);

   sos_Bool result;

   if (get_is_static(_OBSThis) != m.get_is_static())
      result = TRUE;
   else
   {  sos_Param_List pls = get_params(_OBSThis),
		     plm = m.get_params();
      sos_Int        pls_clow, pls_chigh, // minimum and maximum number of
                     plm_clow, plm_chigh; // parameters taking default exprs.
                                          // into account.

      if (INVALID(pls))
         pls_clow = pls_chigh = 0;
      else
      {
         pls_clow = 0;
         agg_iterate (pls, sos_Param p)
            if (VALID(p.get_default_expr()))
               break;
            pls_clow++;
         agg_iterate_end (pls, p)
         pls_chigh = INVALID(pls) ? 0 : pls.card();
      }
      if (INVALID(plm))
         plm_clow = plm_chigh = 0;
      else
      {
         plm_clow = 0;
         agg_iterate (plm, sos_Param p)
            if (VALID(p.get_default_expr()))
               break;
            plm_clow++;
         agg_iterate_end (plm, p)
         plm_chigh = INVALID(plm) ? 0 : plm.card();
      }

      result = (sos_Bool) (pls_chigh < plm_clow OR pls_clow > plm_chigh);
   }
   TT (mta_H, T_LEAVE; TB(result));
   return result;
}


LOCAL sos_Bool mta_check_param_correspondence (const sos_Method& m1,
					       const sos_Type_descr& p1,
					       const sos_Method& m2,
					       const sos_Type_descr& p2,
					       const sos_Method& mgf,
					       const sos_Method& mgfgf,
					       sos_Int pos)
{
   T_PROC ("mta_check_param_correspondence");
   TT (mta_M, T_ENTER);

   /* Checks, if the param type p1 occurring in m1 is `the same' as
      p2 in m2. See the comment in sos_Method::redefines for an explanation
      of `the same'.
      pos denotes the parameter number, where 0 means 'result'.
      When is p2 generic?
      - If p2 is an sos_Gen_param, then it is of course generic
      - If p2 is not an sos_Gen_param, then it must be one in the generic
	class, to which the method of p2 belongs to. Furthermore, the
	generic param, p2 was substituted for, must be instantiated with
	a formal generic param of m1's class. Precondition is thus, that
	m1,get_defined_in() is a generic class.
   */

   sos_Bool result;
   sos_Bool p1_generic, p2_generic;
   sos_Gen_param sub_gp_of_p2;	// in case p2 is generic, and also m1
				// belongs to a generic class, this variable 
				// holds the formal genparam of m1's genclass
				// that instantiated p2.
   sos_Type_descr local_p2 = p2;
   sos_Class_type class_1 = m1.get_defined_in();

   // First check, if we have to cope with genericity for p1:
   p1_generic = p1.has_type (sos_Gen_param_type);
   // Now check, if we have to cope with genericity for p2:
   p2_generic = p2.has_type (sos_Gen_param_type);

   // Only get sos_Gen_param belonging to p2, if m1 is defined in a
   // generic class. Otherwise, m2 must belong to a concrete generic
   // instantiation and hence cannot have a formal gen param of m1's class
   // in its signature.
   if (class_1.is_generic_class())
   {
      sos_Gen_param_List agpl_2 = sos_Gen_param_List::make (NO_OBJECT);
      sos_Class_type gen_class_2;

      if (NOT p2_generic)
      {  // Not a gen param, but might belong to a passed through gen param:
	 // Now we have to get the generic class belonging to m2.
	 sos_Method m2g = sos_Method::make (NO_OBJECT);
	 if (VALID (mgfgf))
	    m2g = mgfgf;
	 else
	    m2g = mgf;
	 if (VALID (m2g))	// was it a method with generic params?
	 {			// Then m2g is defined in a generic class.
	    gen_class_2 = m2g.get_defined_in();
	    if (pos)
	       local_p2 = m2g.get_params().get_nth (pos).get_type();
	    else
	       local_p2 = m2g.get_result_type();
	    if (local_p2.has_type (sos_Gen_param_type))
	    {
	       // Now check, if this genparam was instantiated with a formal
	       // genparam of class_1:
	       agpl_2 = mta_get_agpl_of_superclass (class_1, gen_class_2);
	       sos_Gen_param_List fgpl_2 = gen_class_2.get_formal_gen_params();
	       Index fgpl_pos = fgpl_2.find (sos_Gen_param::make(local_p2));
	       // Now pos contains the position of p2 in the superclass'
	       // formal generic parameter list. Was this a passed through one?
	       sub_gp_of_p2 = agpl_2.get_nth(fgpl_pos);
	       if (VALID (sub_gp_of_p2))	// p2 generic:
		  // params match, if p1 is the corresponding genpar:
		  result = (sos_Bool) (sub_gp_of_p2 == p1);
	       else	// p2 not generic, then p1 must be non-generic too,
	        	// and the make_type()'s must be identical:
		  result = (sos_Bool) (NOT p1_generic AND
			   p1.make_type() == p2.make_type());
	    }
	 }
	 // if m2 had no generic parameters or was no instantiation of a
	 // method in a generic class or p2 was not an instantiated
	 // generic parameter, then the result is TRUE, only if p1 was also
	 // no generic parameter and p1 and p2 have the same type (make_type).
	 if (INVALID (m2g) OR NOT local_p2.has_type (sos_Gen_param_type))
	    result = (sos_Bool) (NOT p1_generic AND
		     p1.make_type() == p2.make_type());
      }
      else 	// p2 is generic, and class_1 is generic, so we can determine
      {		// if and which formal genpar of m1's class instantiated p2.
	 gen_class_2 = m2.get_defined_in();
	 sos_Gen_param_List fgpl_2 = gen_class_2.get_formal_gen_params();
	 Index fgpl_pos = fgpl_2.find (sos_Gen_param::make(local_p2));
	 agpl_2 = mta_get_agpl_of_superclass (class_1, gen_class_2);
	 sub_gp_of_p2 = agpl_2.get_nth (fgpl_pos);
	 // params match, if p1 is the corresponding formal genpar of class_1.
	 result = (sos_Bool) (sub_gp_of_p2 == p1);
      }
      if (VALID(agpl_2))
	 agpl_2.destroy();
   }
   else	// class of m1 was not generic, so p1 is also non-generic and
   {
    	// generic superclasses must be instantiated by concrete types:
      result = (sos_Bool) (p1.make_type() == p2.make_type());
   }
   TT (mta_M, T_LEAVE);
   return result;
}

LOCAL sos_Bool mta_check_resulttype_is_universal (const sos_Method& m1,
					 	  const sos_Type_descr& rt1,
					 	  const sos_Method& m2,
					 	  const sos_Type_descr& rt2)
{
   T_PROC ("mta_check_resulttype_is_universal");
   TT (mta_M, T_ENTER);
   /* Returns TRUE, if the resulttype rt2 of method m2 comes from a
      formal generic parameter of the class where it is defined in,
      and has rt1 as its most general substitution.
   */

   sos_Bool result;
   sos_Type subst_for_rt2;

   if (rt2.has_type (sos_Gen_param_type))
      subst_for_rt2 = rt2.make_type();
   else
   {
      sos_Method m2g = m2.get_generated_from();
      if (INVALID (m2g))
	 m2g = m2;	// was not generic method at all, or generic class
      else
	 if (INVALID (m2g.get_generated_from()))
	 // m2g is generic method, but not root. Go back one step:
	    m2g = m2;
      subst_for_rt2 = m2g.get_result_type().make_type();
   }
   result = (sos_Bool) (subst_for_rt2 == rt1.make_type());

   TT (mta_M, T_LEAVE);
   return result;
}


// *************************************************************************
sos_Bool _sos_Method::redefines (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Method) m)
// *************************************************************************
// precondition: self.get_name() == m.get_name()

const{  T_PROC ("sos_Method::redefines")
   TT (mta_H, T_ENTER);

   // PRECONDITION: m.get_defined_in() dominates self.get_defined_in()

   /* Attention must be paid to the signature comparison in case self
      has generic parameters in its signature or m is generated_from
      a method with generic parameters:
      If we take make_type() of every parameter, then we replace all
      generic parameters by their most general substitution. This is the type,
      we would expect arguments of methods to be of, which were originally
      the passed through generic parameter. We could not detect any more, if
      one of the two parameters was not a generic parameter, which would cause
      only one substitution for that parameter to be correct.
      If we do the complicated check in case self's parameter
      is a generic parameter, if the superclass method's parameter 
      is actually the passed through generic parameter, then we could
      issue a message in case this test fails, because the correctness of
      the redefinition then depends on the actual generic parameter used
      for the subclass.
      If a parameter type in self is concrete (not a generic param), then
      we have to leave this unchanged, regardless if this concrete type is
      a substitute for a generic param. We are considering a concrete
      instantiated parameter.
      For m we have to distinguish, because parameter types will always be
      concrete (we never have generic classes as superclasses; these are
      always instantiations (probably with formal generic parameters of the
      subclass)), so this doesn't let us know, if the type comes from the
      most general restriction of a generic parameter, or if the type was
      provided explicitly:
      We can check, if m's generated_from() leads (transitively) to a generic
      class. In this case we can determine, if the corresponding parameter
      in the generic class was a generic parameter. If not, then it really
      was a concrete type. Otherwise, we have to examine, if the genparam
      used as method param was passed through from the generic class
      belonging to self, and there was used as the corresponding parameter
      of self. If self was not declared in a generic class
      (we get this by following the generated_from() paths of self), then
      this is an error, since the user tried to redefine a generic method
      by a concrete method.
	  - In the current version, the genCC doesn't cope with baseclasses
	    not being a root class, if the generic parameter differing from
	    the instantiation in the root class is used as result of the
	    method under consideration.
	    Therefore, the user must use the most general type for the
	    result type, which we must explicitely allow here. This must
	    be forbiddden in the final version, when genCC can handle this
	    (by generating an implementation class for the non-root
	    superclass).
   */

   sos_Bool	  result;
   sos_Param_List pls = get_params(_OBSThis),
		  plm =    m.get_params();
   sos_Method     mgf = m.get_generated_from(),
		  mgfgf = VALID (mgf) ? mgf.get_generated_from()
				      : sos_Method::make (NO_OBJECT);

   if ((INVALID(pls) != INVALID(plm))
       || (get_is_static(_OBSThis) != m.get_is_static()))
      result = FALSE;
   else
   {  sos_Type_descr tds = get_result_type(_OBSThis);
      sos_Type_descr tdm = m.get_result_type();
      sos_Bool result_ok = mta_check_param_correspondence (sos_Method::make(_OBSThis,this), tds, m,
							   tdm, mgf, mgfgf, 0);
      if (NOT result_ok)	// do we have constrained generic params?
	 result_ok = mta_check_resulttype_is_universal (sos_Method::make(_OBSThis,this), tds, m, tdm);
      if (result_ok)
      {  result = TRUE;
	 if VALID (pls) 
         {  int comp;
            sos_Int pos_count = 0;
            agg_iterate_double (pls, sos_Param ps, plm, sos_Param pm, comp)
            {  pos_count++;
               tds = ps.get_type();
               tdm = pm.get_type();
               sos_Expr des = ps.get_default_expr();
               sos_Expr dem = pm.get_default_expr();
               result = (sos_Bool)(
			   mta_check_param_correspondence (sos_Method::make(_OBSThis,this), tds, m, 
							   tdm, mgf, mgfgf,
							   pos_count)
                           AND (INVALID(dem)
                           OR (VALID(des) AND des.equal(dem)))); // NO_OBJECT!
               if (NOT result)
                  break;
            }
            agg_iterate_double_end (pls, ps, plm, pm, comp);
            result = (sos_Bool)(result AND
                     (comp == 0 OR (comp > 0
                     AND VALID(pls.get_nth(pos_count+1).get_default_expr()))));
	 }
      }
      else
         result = FALSE;
   }
   TT (mta_H, T_LEAVE; TB(result));
   return result;
}

// *************************************************************************
sos_Object _sos_Method::execute (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Object) o,OBST_PARDECL( sos_Object_Array) p)
// *************************************************************************
const{  T_PROC ("sos_Method::execute")
   TT (mta_H, T_ENTER);

   sos_Object result;
   sos_Bool   found = FALSE;

   sos_Method_impl_List impls = get_impls(_OBSThis);
   if VALID (impls) 
   {  agg_iterate (impls, sos_Method_impl impl)
         if (impl.isa (cci_Method_impl_type))
	 {  result = cci_Method_impl::make (impl).execute (o, p);
	    found = TRUE;
	    break;
	 }
      agg_iterate_end (impls, impl);
   }
   if (NOT found)
      err_raise (err_SYS, err_MTA_METHOD_NOT_EXECUTABLE,
			  get_name(_OBSThis).make_Cstring());
   TT (mta_H, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Method_table
// -------------------------------------------------------------------------
// *************************************************************************
sos_Method _sos_Method_table::lookup (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Method) m)
// *************************************************************************
const{  T_PROC ("sos_Method_table::lookup")
   TT (mta_H, T_ENTER);

   sos_Method result = sos_Method::make (NO_OBJECT);
   sos_String m_name = m.get_name();

   if (is_key (_OBSThis,m_name))
   {  sos_Method_List ml = sos_Method_table::make(_OBSThis,this)[m_name];

      agg_iterate (ml, sos_Method m1)
         if (NOT m.overloads (m1))
	 {  result = m1;
	    break;
	 }
      agg_iterate_end (ml, m1);
   }
   TT (mta_H, T_LEAVE);
   return result;
}

// *************************************************************************
sos_Method _sos_Method_table::lookup_or_add (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Method) m)
// *************************************************************************
const{  T_PROC ("sos_Method_table::lookup_or_add")
   TT (mta_H, T_ENTER);

   sos_Method result;
   sos_Bool   found = FALSE;

   sos_String      m_name = m.get_name();
   sos_Method_List ml     = sos_Method_table::make(_OBSThis,this)[m_name];
   if INVALID (ml) 
   {  ml = sos_Method_List::create (_OBSThis.container(), FALSE);
      sos_String m_key = sos_String::clone (m_name, _OBSThis.container());
      insert (_OBSThis,m_key, ml);
   }
   else
   {  agg_iterate (ml, sos_Method m1)
         if (NOT m.overloads (m1))
	 {  found = TRUE;
	    result = m1;
	    break;
	 }
      agg_iterate_end (ml, m1);
   }
   if (NOT found)
   {  ml.append (m);
      result = m;
   }
   TT (mta_H, T_LEAVE; TB(found));
   return result;
}

// *************************************************************************
sos_Method _sos_Method_table::replace_or_add (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Method) m)
// *************************************************************************
const{  T_PROC ("sos_Method_table::replace_or_add")
   TT (mta_H, T_ENTER);

   sos_Method result;
   sos_Bool   found = FALSE;

   sos_String      m_name = m.get_name();
   sos_Method_List ml     = sos_Method_table::make(_OBSThis,this)[m_name];
   if INVALID (ml) 
   {  ml = sos_Method_List::create (_OBSThis.container(), FALSE);
      sos_String m_key = sos_String::clone (m_name, _OBSThis.container());
      insert (_OBSThis,m_key, ml);
   }
   else
   {  sos_Cursor c = ml.open_cursor();
      for (sos_Bool valid = ml.is_valid (c);  valid;  valid = ml.to_succ (c))
      {  sos_Method m1 = ml.get (c);
	 if (NOT m.overloads (m1))
	 {  ml.set (c, m);
	    found  = TRUE;
	    result = m1;
	    break;
	 }
      }
      ml.close_cursor (c);
   }
   if (NOT found)
   {  ml.append (m);
      result = m;
   }
   TT (mta_H, T_LEAVE);
   return result;
}

// *************************************************************************
sos_Method _sos_Method_table::lookup_comp (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_String) name,
					  sos_Bool   is_set)
// *************************************************************************
const{  T_PROC ("sos_Method_table::lookup_comp")
   TT (mta_H, T_ENTER);

   sos_Method result = sos_Method::make (NO_OBJECT);

   sos_String mn = _sos_String::create (TEMP_CONTAINER,
				       (is_set ? "set_" : "get_"));
   mn += name;

   sos_Method_List ml = sos_Method_table::make(_OBSThis,this)[mn];
   if VALID (ml) 
   {  agg_iterate (ml, sos_Method m)
         if VALID (m.get_comp_descr() )// component method found?
	 {  result = m;
	    break;
	 }
      agg_iterate_end (ml, m);
   }
   mn.destroy();

   TT (mta_H, T_LEAVE);
   return result;
}

// *************************************************************************
void _sos_Method_table::remove_method (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Method) m)
// *************************************************************************
const{  T_PROC ("sos_Method_table::remove_method")
   TT (mta_H, T_ENTER);

   sos_String m_name = m.get_name();

   if (is_key (_OBSThis,m_name))
   {  sos_Method_List ml = sos_Method_table::make(_OBSThis,this)[m_name];

      Index i = 1;
      agg_iterate (ml, sos_Method m1)
         if (NOT m.overloads (m1))
         {  ml.remove (i);
            break;
         }
	 else
	    i++;
      agg_iterate_end (ml, m1);

      if (ml.card() == 0)  // no more method with that name
      {			   // Then remove that entry from the mapping:
	 sos_Cursor c = open_cursor(_OBSThis);
	 move_cursor (_OBSThis,c, m_name);
	 sos_Object m_key = get_key (_OBSThis,c);
	 remove (_OBSThis,m_name); // remove that list from the mapping
	 m_key.destroy();      // The string was cloned during insert
#ifdef OBST_PROT_ALL
	 sos_ModifHistory::trash (ml);
#else
	 ml.destroy();
#endif
      }
   }
   
   TT (mta_H, T_LEAVE);
}


// -------------------------------------------------------------------------
// sos_Type_descr
// -------------------------------------------------------------------------
// *************************************************************************
sos_Type _sos_Type_descr::make_type (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  // default implementation expects self to be derived from sos_Type.
   T_PROC("sos_Type_descr::make_type")
   TT (mta_H, T_ENTER);

   sos_Type result = sos_Type::make(sos_Type_descr::make(_OBSThis,this));

   TT (mta_H, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Generic_instantiation
// -------------------------------------------------------------------------
// *************************************************************************
sos_Type _sos_Generic_instantiation::make_type (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Generic_instantiation::make_type")
   TT (mta_H, T_ENTER);

   sos_Type result = get_instantiation(_OBSThis);

   TT (mta_H, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Super_class
// -------------------------------------------------------------------------
// *************************************************************************
sos_Type _sos_Super_class::make_type (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Super_class::make_type")
   TT (mta_H, T_ENTER);

   sos_Type result = get_super_class(_OBSThis).make_type();

   TT (mta_H, T_LEAVE);
   return result;
}


// -------------------------------------------------------------------------
// sos_Gen_param
// -------------------------------------------------------------------------
// *************************************************************************
sos_Type _sos_Gen_param::make_type (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Gen_param::make_type")
   TT (mta_H, T_ENTER);

   sos_Type result = INVALID(get_super_type(_OBSThis))
		       ? sos_Type::make(_obst_knl_type(_THE_OBJECT_TYPE))
		       : get_super_type(_OBSThis).make_type();

   TT (mta_H, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Enum_type
// -------------------------------------------------------------------------
// *************************************************************************
sos_Scalar_object _sos_Enum_type::make_object (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_String) literal)
// *************************************************************************
const{  T_PROC ("sos_Enum_type::make_object")
   TT (mta_H, T_ENTER);

   sos_String_List sl    = get_literals(_OBSThis);
   sos_Int         idx   = 0;
   sos_Bool        found = FALSE;

   agg_iterate (sl, sos_String lit)
      if (lit.equal (literal))
      {  found = TRUE;
         break;
      }
      ++ idx;
   agg_iterate_end (sl, lit);

   sos_Scalar_object result = found ? _obst_enum_from_sos_Int (idx, (sos_Type)sos_Enum_type::make(_OBSThis,this))
				    : sos_Scalar_object::make (NO_OBJECT);
   TT (mta_H, T_LEAVE);
   return result;
}

// *************************************************************************
sos_String _sos_Enum_type::make_string (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Scalar_object) enum_obj)
// *************************************************************************
const{  T_PROC ("sos_Enum_type::make_string")
   TT (mta_H, T_ENTER);

   int    	   lidx     = _obst_Int_from_enum ((sos_Object)enum_obj, (sos_Type)sos_Enum_type::make(_OBSThis,this)) + 1;
   sos_String_List literals = get_literals(_OBSThis);

   sos_String result
      = (1 <= lidx  AND  lidx <= literals.card())
	   ? sos_String::clone (literals.get_nth (lidx), TEMP_CONTAINER)
	   : sos_String::make (NO_OBJECT);

   TT (mta_H, T_LEAVE);
   return result;
}

// *************************************************************************
sos_Bool _sos_Enum_type::is_scalar (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Enum_type::is_scalar")
   TT (mta_H, T_ENTER);
   TT (mta_H, T_LEAVE);

   return TRUE;
}

// *************************************************************************
sos_Bool _sos_Enum_type::is_derived_from (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Type) tp)
// *************************************************************************
// note: is_derived_from_some is implemented in sos_Type as is_derived_from.
//       This relies on the fact that both relations are equivalent for
//	 sos_Scalar_object.

const{  T_PROC ("sos_Enum_type::is_derived_from")
   TT (mta_H, T_ENTER);

   sos_Bool result;

   if (_OBSThis.get_id() == tp._self_id()
#ifdef BOOT
       OR _obst_knl_type(_THE_SCALAR_TYPE).is_derived_from (tp)
#else
       OR sos_Scalar_object_type.is_derived_from (tp)
#endif
      )
      result = TRUE;

   else if (tp._type_id() == _sos_Union_type_type)
      result = sos_Union_type::make(tp).is_base_of (sos_Enum_type::make(_OBSThis,this));

   else
      result = FALSE;

   TT (mta_H, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Extern_type
// -------------------------------------------------------------------------
// *************************************************************************
sos_Bool _sos_Extern_type::is_scalar (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Extern_type::is_scalar")
   TT (mta_H, T_ENTER);
   TT (mta_H, T_LEAVE);

   return TRUE;
}

// *************************************************************************
sos_Bool _sos_Extern_type::is_derived_from (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Type) tp)
// *************************************************************************
// note: is_derived_from_some is implemented in sos_Type as is_derived_from.
//       This relies on the fact that both relations are equivalent for
//	 sos_Scalar_object.

const{  T_PROC ("sos_Extern_type::is_derived_from")
   TT (mta_H, T_ENTER);

   sos_Bool result;

   if (_OBSThis.get_id() == tp._self_id()
#ifdef BOOT
       OR _obst_knl_type(_THE_SCALAR_TYPE).is_derived_from (tp)
#else
       OR sos_Scalar_object_type.is_derived_from (tp)
#endif
      )
      result = TRUE;

   else if (tp._type_id() == _sos_Union_type_type)
      result = sos_Union_type::make(tp).is_base_of (sos_Extern_type::make(_OBSThis,this));

   else
      result = FALSE;

   TT (mta_H, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Union_type
// -------------------------------------------------------------------------
// *************************************************************************
sos_Bool _sos_Union_type::is_base_of (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Type) tp)
// *************************************************************************
const{  T_PROC ("sos_Union_type::is_base_of")
   TT (mta_H, T_ENTER);

   sos_Bool result  = FALSE;

   sos_Type_List tl = get_united(_OBSThis);
   agg_iterate (tl, sos_Type utp)
      if (tp.is_derived_from (utp))
      {  result = TRUE;
	 break;
      }
   agg_iterate_end (tl, utp);

   TT (mta_H, T_LEAVE);
   return result;
}

// -------------------------------------------------------------------------
// sos_Typedef_type
// -------------------------------------------------------------------------
// *************************************************************************
sos_Type _sos_Typedef_type::make_type (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Typedef_type::make_type")
   TT (mta_H, T_ENTER);

   sos_Type t = get_named_type(_OBSThis);

   TT (mta_H, T_LEAVE);
   return t;
}

// -------------------------------------------------------------------------
// sos_Class_type
// -------------------------------------------------------------------------
// *************************************************************************
sos_Bool _sos_Class_type::is_generic_class (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Class_type::is_generic_class")
   TT(mta_H, T_ENTER);

   sos_Bool result = VALID (get_formal_gen_params(_OBSThis)) ;

   TT(mta_H, T_LEAVE);
   return result;
}

// *************************************************************************
sos_Bool _sos_Class_type::is_instantiation (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Class_type::is_instantiation")
   TT(mta_H, T_ENTER);

   sos_Bool result = VALID (get_actual_gen_params(_OBSThis)); 

   TT(mta_H, T_LEAVE);
   return result;
}

// *************************************************************************
sos_Type _sos_Class_type::generic_param (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Class_type) rtp, sos_Int n)
// *************************************************************************
const{  T_PROC ("sos_Class_type::generic_param")
   TT(mta_H, T_ENTER);

   //PRE: self.is_derived_from_some (rtp) &&
   //     self.is_instantiation() && rtp.is_instantiation()

   sos_Type result = sos_Type::make(NO_OBJECT);

   sos_Super_class_List scl = get_super_closure(_OBSThis);
   agg_iterate_reverse (scl, sos_Super_class sc)
      sos_Class_type sctp = sos_Class_type::make (sc.make_type());
      if (sctp.root() EQ rtp)
      {  result = sctp.get_actual_gen_params().get_nth (n).make_type();
	 break;
      }
   agg_iterate_reverse_end (scl, sc);

   TT(mta_H, T_LEAVE);
   return result;
}

// *************************************************************************
sos_Type _sos_Class_type::root (const sos_Typed_id&_OBSThis)
// *************************************************************************
const{  T_PROC ("sos_Class_type::root")
   TT (mta_H, T_ENTER);

   sos_Type tp = get_root_class(_OBSThis);

   TT (mta_H, T_LEAVE);
   return tp;
}

// *************************************************************************
sos_Bool _sos_Class_type::is_derived_from (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Type) tp)
// *************************************************************************
// pre: self.make_type() == self
// method should maybe be implemented by a CCI hashtable lookup.

const{  T_PROC ("sos_Class_type::is_derived_from")
   TT (mta_H, T_ENTER);

   sos_Bool result  = FALSE;

   if (   tp._self_id() == _OBSThis.get_id()
#ifdef BOOT
       OR tp._self_id() == _obst_knl_type (_THE_OBJECT_TYPE)._self_id()
#else
       OR tp._self_id() == _sos_Object_type
#endif
      )
      result = TRUE;

   else if (tp._type_id() == _sos_Union_type_type)
      result = sos_Union_type::make(tp).is_base_of (sos_Class_type::make(_OBSThis,this));

   else
   {  sos_Super_class_List scl = get_super_closure(_OBSThis);

      agg_iterate (scl, sos_Super_class sc)
         if (sc.get_super_class() ==  tp)
	 {  result = TRUE;
	    break;
	 }
      agg_iterate_end (scl, sc);
   }
   TT (mta_H, T_LEAVE; TB (result));
   return result;
}

// *************************************************************************
sos_Bool _sos_Class_type::is_derived_from_some (const sos_Typed_id&_OBSThis,OBST_PARDECL(sos_Type) tp)
// *************************************************************************
// pre: self.make_type() == self

const{  T_PROC ("sos_Class_type::is_derived_from_some")
   TT (mta_H, T_ENTER);

   sos_Bool result  = FALSE;

   if (   tp._self_id() == _OBSThis.get_id()
#ifdef BOOT
       OR tp._self_id() == _obst_knl_type (_THE_OBJECT_TYPE)._self_id()
#else
       OR tp._self_id() == _sos_Object_type
#endif
      )
      result = TRUE;

   else if (tp._type_id() == _sos_Union_type_type)
      result = sos_Union_type::make(tp).is_base_of (sos_Class_type::make(_OBSThis,this));

   else
   {  sos_Super_class_List scl = get_super_closure(_OBSThis);

      agg_iterate (scl, sos_Super_class sc)
         if (sos_Class_type::make(sc.get_super_class()).root() == tp)
	 {  result = TRUE;
	    break;
	 }
      agg_iterate_end (scl, sc);
   }
   TT (mta_H, T_LEAVE; TB (result));
   return result;
}
