#ifndef _OBST_STATS_H
#define _OBST_STATS_H 1
/* --------------------------------------------------------------------------
 * Copyright 1992-1994 by Forschungszentrum Informatik (FZI)
 * All rights reserved.
 *
 * You can use and distribute this software under the terms of the license
 * you should have received along with this software; either version 1.1 of
 * the license, or (at your option) any later version.
 * For a copy of the license or for additional information about this software,
 * write to Xcc Software, Durlacher Allee 53, D-76131 Karlsruhe, Germany;
 * Email: obst@xcc-ka.de.
 * --------------------------------------------------------------------------
 */

typedef struct
{  int occurences;		/* number of times this 'data' occurs	*/
   int data;			/* counted data entry			*/
} obst_histentry;

typedef struct
{  char*	   descr;	/* name of hash table (must not be freed)    */
   int		   buckets;	/* number of hash buckets		     */
   int		   histsize;	/* size of 'histogram' array		     */
   obst_histentry* histogram;   /* histogram of bucket sizes (must be freed) */
} obst_htabinfo;

typedef struct
{  obst_htabinfo*  htabs;	/* array of hash table statistics	*/
   int		   size;	/* size of 'htabs'			*/
} obst_htabstat;

/*
 * Builds up the histogram array '*h' of size '*size' by counting the
 * occurences of 'data' values.
 * The array is dynamically allocated and should be freed by the caller.
 * The first call must be done with '*size == 0'.
 * The array will be ordered by ascending 'data' values.
 */
extern void obst_add_histdata __FSIG_((obst_histentry** h, int* sz, int data));

/*
 * Adds an entry to the 'htabs' field and returns it. This entry will be
 * dynamically allocated.
 *
 * WARNING: Existing pointers into the 'htabs' field will in general become
 *	    invalid.
 */
extern obst_htabinfo* obst_add_htabinfo __FSIG_((obst_htabstat*));

/*
 * Frees the space associated with the given statistics array which must have
 * been built up via the _obst_add* functions.
 * 'htstat->size' must be 0 for the first call.
 */
extern void obst_free_htabstat __FSIG_((obst_htabstat* htstat));

/*
 * Collects the available hash table statistics, 'htstat' is fully initialized.
 */
extern void obst_collect_stats __FSIG_((obst_htabstat* htstat));
extern void _cci_collect_stats __FSIG_((obst_htabstat*));
extern void _psm_collect_stats __FSIG_((obst_htabstat*));

/*
 * Generate a text from the given hash table statistics. 'htstat' is cleared
 * afterwards if 'del' is set.
 * The result string must be deleted by the caller.
 */
#include "knl_use.h"
extern char* obst_print_htabstats __FSIG_((obst_htabstat* htstat,
					   sos_Bool       del));
#endif
