/* --------------------------------------------------------------------------
 * Copyright 1992-1994 by Forschungszentrum Informatik (FZI)
 * All rights reserved.
 *
 * You can use and distribute this software under the terms of the license
 * you should have received along with this program.
 * For a copy of the license or for additional information about this software,
 * write to Xcc Software, Durlacher Allee 53, D-76131 Karlsruhe, Germany;
 * Email: obst@xcc-ka.de.
 * --------------------------------------------------------------------------
 */
/* OBST LIBRARY MODULE */
// **************************************************************************
// Module prot                   13-Aug-1993	       Bernhard Schiefer (bs)
// **************************************************************************
// implements methods of all classes declared in schema "prot"
// **************************************************************************
 
#include "obst_progstd.h"
#include "obst.h"
#include "smg.h"
#include "mta_obst.h"
#include "trc_prot.h"

#include "prot_obst.h"
#include "prot_err.h"

// -------------------------------------------------------------------------
// sos_ClassModif
// -------------------------------------------------------------------------
// *************************************************************************
void sos_ClassModif::local_initialize(sos_ClassModif cm)
// *************************************************************************
{
   T_PROC("sos_ClassModif::local_initialize");
   TT(prot_M, T_ENTER);

   sos_Container cnt = cm.container();

   cm.set_change_create_params(sos_ParamModif_List::create(cnt));

   cm.set_added_friends         (sos_Type_Set::create(cnt));
   cm.set_deleted_friends       (sos_Type_Set::create(cnt));
  
   cm.set_change_superclasses   (sos_SuperClassModif_List::create(cnt));

   cm.set_change_comps          (sos_CompModif_List::create(cnt));

   cm.set_old_comp_offset(sos_Comp_descr_sos_Offset_Mapping::create(cnt));
   cm.set_old_spcl_offset(sos_Type_descr_sos_Offset_Mapping::create(cnt));
 
   cm.set_old_local_size(-1);
   cm.set_old_object_size(-1);
 
   cm.set_added_methods         (sos_Method_Set::create(cnt));
   cm.set_deleted_methods       (sos_Method_Set::create(cnt));
   cm.set_modified_methods      (sos_MethodModif_List::create(cnt));

   TT(prot_M, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::local_finalize(sos_ClassModif cm)
// *************************************************************************
{
   T_PROC("sos_ClassModif::local_finalize");
   TT(prot_M, T_ENTER);

   sos_ParamModif_List             pm_list;
   sos_ParamModif                  pm;
   sos_SuperClassModif_List        scm_list;
   sos_SuperClassModif             scm;
   sos_CompModif_List              cpm_list;
   sos_CompModif                   cpm;
   sos_MethodModif_List            mm_list;
   sos_MethodModif                 mm;
   

   cm.deinstall();

   pm_list = cm.get_change_create_params();
   agg_iterate(pm_list, pm)
   {
      pm.destroy();
   } agg_iterate_end(pm_list, pm);
   pm_list.destroy();

   cm.get_added_friends().destroy();
   cm.get_deleted_friends().destroy();

   scm_list = cm.get_change_superclasses();
   agg_iterate(scm_list, scm)
   {
      scm.destroy();
   } agg_iterate_end(scm_list, scm);
   scm_list.destroy();

   cpm_list = cm.get_change_comps();
   agg_iterate(cpm_list, cpm)
   {
      cpm.destroy();
   } agg_iterate_end(cpm_list, cpm);
   cpm_list.destroy();

   cm.get_added_methods().destroy();
   cm.get_deleted_methods().destroy();
  
   mm_list = cm.get_modified_methods();
   agg_iterate(mm_list, mm)
   {
      mm.destroy();
   } agg_iterate_end(mm_list, mm);
   mm_list.destroy();

   cm.get_old_comp_offset().clear();
   cm.get_old_comp_offset().destroy();
   cm.get_old_spcl_offset().clear();
   cm.get_old_spcl_offset().destroy();

   TT(prot_M, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::install(sos_Schema_module sm)
// *************************************************************************
{
   T_PROC("sos_ClassModif::install");
   TT(prot_M, T_ENTER);

   sos_SchemaHistory sh = sos_ModifHistory::modif_history().get_history()[sm];
   sos_Class_type    ct;
   sos_VCM_List      vcm;

   ct  = sos_Class_type::make(self.get_modified_type()); 
   vcm = sh.get_class_modifs()[ct];
   
   if (INVALID(vcm))
   {  vcm = sos_VCM_List::create(self.container());
      sh.get_class_modifs().insert(ct, vcm);
   }

   vcm.insert(1, self);

   TT(prot_M, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::deinstall()
// *************************************************************************
{
   T_PROC("sos_ClassModif::deinstall");
   TT(prot_M, T_ENTER);

   sos_SchemaHistory sh;
   sos_Schema_module sc;
   sos_Class_type    ct;
   Index             pos;
   sos_VCM_List      vl;
 
   ct = sos_Class_type::make(self.get_modified_type());
   sc = sos_Schema_module::retrieve(ct.container()); 
   if (VALID(sc))
   {
     sh = sos_ModifHistory::modif_history().get_history()[sc]; 
     if (VALID(sh))
     {  vl = sh.get_class_modifs()[ct];
        if (VALID(vl))
        {  pos = vl.find(self);
           vl.remove(pos);
           if (vl.is_empty()) 
             sh.get_class_modifs().remove(ct);
        }
     }
   }
   TT(prot_M, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::save_offsets(sos_Bool only_spcl)
// *************************************************************************
{
   T_PROC("sos_ClassModif::save_offsets");
   TT(prot_M, T_ENTER);

   sos_Schema_module    sc;
   sos_SchemaModif      sm;
   sos_Class_type_List  subclasses;
   sos_Class_type       sub;
   sos_ClassModif       subcm;	
   sos_Super_class_List superclasses;
   sos_Super_class      super;
   sos_Comp_descr_List  components;
   sos_Comp_descr	comp;
   sos_Class_type       ct;

   ct         = sos_Class_type::make(self.get_modified_type());
   subclasses = ct.get_subclasses();

   // save the superclasses offsets for all subclasses and the class itself,
   // if they have not already been saved
   agg_iterate(subclasses, sub)
   {
      if (sub != ct)
      {  sc = sos_Schema_module::retrieve(sub.container());
         sm = sos_SchemaModif::mod_schema(sc); 
 
         subcm = sos_ClassModif::make(sm.mod_type(sub));
      }
      else
         subcm = self;

      if (VALID(subcm) && subcm.get_old_spcl_offset().is_empty())
      {  superclasses = sub.get_super_closure();
    
         agg_iterate(superclasses, super) 
         {  
            subcm.get_old_spcl_offset().insert(super.get_super_class(),
					       super.get_offset());
         } agg_iterate_end(superclasses, super);
  
         subcm.set_old_local_size (sub.get_local_size());
         subcm.set_old_object_size(sub.get_object_size()); 
      }

   } agg_iterate_end(subclasses, sub);
  

   // save the relative component offsets of the class
   // if they have not already been saved
   if (!only_spcl && self.get_old_comp_offset().is_empty())
   {
      components = ct.get_components();
 
      agg_iterate(components, comp)
      {
         self.get_old_comp_offset().insert(comp, comp.get_offset());
      } agg_iterate_end(components, comp);
   }
   TT(prot_M, T_LEAVE);
} 

// *************************************************************************
void sos_ClassModif::add_create_par(sos_Param param, sos_Int position)
// *************************************************************************
{
   T_PROC("sos_ClassModif::add_create_par");
   TT(prot_H, T_ENTER);

   sos_ParamModif pm;
 
   pm = sos_ParamModif::create(self.container());
   pm.set_modified_param(param);
   pm.set_position      (position);
   pm.set_change        (PROT_ADD);
 
   self.get_change_create_params().append(pm);

   self.set_change(PROT_MOD);

   TT(prot_H, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::del_create_par(sos_Param param)
// *************************************************************************
{
   T_PROC("sos_ClassModif::del_create_par");
   TT(prot_H, T_ENTER);

   sos_ParamModif       pm;
   sos_Class_type       ct;

   ct = sos_Class_type::make(self.get_modified_type()); 

   pm = sos_ParamModif::create(self.container());
   pm.set_modified_param(param);
   pm.set_position      (ct.get_create_params().find(param));
   pm.set_change        (PROT_DEL);
 
   self.get_change_create_params().append(pm);

   self.set_change(PROT_MOD);

   TT(prot_H, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::mod_create_par(sos_Param param)
// *************************************************************************
{
   T_PROC("os_ClassModif::mod_create_par");
   TT(prot_H, T_ENTER);

   sos_Container        cnt;
   sos_ParamModif_List  parammodifs;
   sos_ParamModif       pmod;
   sos_ParamModif       pm;
   sos_Class_type       ct;
   sos_Bool             found = FALSE;
 
   parammodifs = self.get_change_create_params();
 
   agg_iterate(parammodifs, pmod)
   {
      if (pmod.get_modified_param() == param)
      {  found = TRUE;
         pm    = pmod;
      }
   } agg_iterate_end(parammodifs, pmod);
 
 
   if (found)
   {  if (pm.get_change() == PROT_DEL)
      {  // param is deleted
         err_raise(err_SYS, err_PROT_PARAM_DEL, prot_ERR, FALSE);
      }
   }
   else
   {
      ct   = sos_Class_type::make(self.get_modified_type());
      cnt  = self.container();
      pmod = sos_ParamModif::create(cnt);
      pmod.set_modified_param(param);
      pmod.set_position      (ct.get_create_params().find(param));
      pmod.set_change        (PROT_MOD);
 
      pmod.set_old_name(sos_String::clone(param.get_name(), cnt));
      pmod.set_old_type      (param.get_type());
      pmod.set_old_default   (param.get_default_expr());
      pmod.set_old_ref       (param.get_is_ref());
 
      parammodifs.append(pmod);

      self.set_change(PROT_MOD);
   }
   TT(prot_H, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::add_super(sos_Super_class spcl)
// *************************************************************************
{
   T_PROC("sos_ClassModif::add_super");
   TT(prot_H, T_ENTER);

   sos_SuperClassModif_List supermodifs;  
   sos_SuperClassModif      spmod;
   sos_SuperClassModif      spm;
   sos_Super_class          super;
   sos_Bool                 found = FALSE;

   supermodifs = self.get_change_superclasses();


   // search whether superclass has been touched before 
   agg_iterate(supermodifs, spmod)
   {
      super = spmod.get_super_class();
      if (!found &&
          (super == spcl  ||  super.make_type() == spcl.make_type()))
      {  // corresponding superclass is found 
         found      = TRUE;
         spm        = spmod;
      }
   } agg_iterate_end(supermodifs, spmod);


   // evaluate search
   if (found)   
   {  if (spm.get_change() == PROT_DEL)
      {  // superclass deleted: mark it as modified and insert new spcl-OID as 
         // new_super_class, just in case the OID of the superclass has changed
         spm.set_change(PROT_MOD);
         spm.set_new_super_class(spcl); 
      }
      else
      {  // superclass modified: impossible, error in MTA 
         // superclass added   : impossible, error in MTA
         err_raise(err_SYS, err_PROT_SUPERCLASS_ADD, prot_ERR, FALSE); 
      }
   }
   else
   {  // superclass not found: insert added superclass 
      spmod = sos_SuperClassModif::create(self.container());
      spmod.set_super_class    (spcl);
      spmod.set_change         (PROT_ADD);
      spmod.set_new_super_class(sos_Super_class::make(NO_OBJECT));

      supermodifs.append(spmod);

      // if found true, the offsets have been saved before
      self.save_offsets(TRUE);

      self.set_change(PROT_MOD);
   }
   TT(prot_H, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::del_super(sos_Super_class spcl)
// *************************************************************************
{
   T_PROC("sos_ClassModif::del_super");
   TT(prot_H, T_ENTER);

   sos_SuperClassModif_List supermodifs;
   sos_SuperClassModif      spmod;
   sos_SuperClassModif      spm;
   sos_Bool                 found  = FALSE;
   sos_Int                  index  = 0;
   sos_Int                  spm_index;
 
   supermodifs = self.get_change_superclasses();
 

   // search whether superclass was touched before
   agg_iterate(supermodifs, spmod)      
   {
      index ++;
 
      if (!found) 
      {  if (   spmod.get_super_class()      == spcl
	     || spmod.get_new_super_class()  == spcl)
         {  found      = TRUE;
            spm        = spmod; 
            spm_index  = index;
         }
      }
   } agg_iterate_end(supermodifs, spmod);


 
   // evaluate search 
   if (found)
   {  if (spm.get_change() == PROT_ADD)
      {  // superclass has been added before: remove modif
         supermodifs.remove(spm_index);
         spm.destroy();
      }
      else if (spm.get_change() == PROT_MOD)
      {  // superclass has been deleted and added before: change modif
         spm.set_change(PROT_DEL);
         spm.set_new_super_class(sos_Super_class::make(NO_OBJECT));
      }
      else 
      {  // superclass has been deleted before
         err_raise(err_SYS, err_PROT_SUPERCLASS_DEL, prot_ERR, FALSE); 
      }
   } 
   else
   {  // superclass not found: insert deleted superclass    
      spmod = sos_SuperClassModif::create(self.container()); 
      spmod.set_super_class    (spcl);
      spmod.set_change         (PROT_DEL);
      spmod.set_new_super_class(sos_Super_class::make(NO_OBJECT));
     
      supermodifs.append(spmod);
 
      // if found true, the offsets have been saved before
      self.save_offsets(TRUE);
    
      self.set_change(PROT_MOD);
   }
   TT(prot_H, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::add_friend(sos_Type fd) 
// *************************************************************************
{
   T_PROC("sos_ClassModif::add_friend");
   TT(prot_H, T_ENTER);

   if (! fd.has_type(sos_Class_type_type))
   {  // friend is not a class
      err_raise(err_SYS, err_PROT_FRIEND_NO_CLASS, prot_ERR, FALSE);
   }

   if (self.get_deleted_friends().is_element(fd))
      self.get_deleted_friends().remove(fd);
   else
   {  self.get_added_friends().insert(fd);
      self.set_change(PROT_MOD);
   }
   TT(prot_H, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::del_friend(sos_Type fd) 
// *************************************************************************
{
   T_PROC("sos_ClassModif::del_friend");
   TT(prot_H, T_ENTER);

   if (!fd.has_type(sos_Class_type_type))
   {  // friend is not a class
      err_raise(err_SYS, err_PROT_FRIEND_NO_CLASS, prot_ERR, FALSE); 
   }
 
   if (self.get_added_friends().is_element(fd)) 
      self.get_added_friends().remove(fd);
   else
   {  self.get_deleted_friends().insert(fd);
      self.set_change(PROT_MOD);
   }
   TT(prot_H, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::add_comp(sos_Comp_descr comp)
// *************************************************************************
{
   T_PROC("sos_ClassModif::add_comp");
   TT(prot_H, T_ENTER);

   sos_CompModif       cpad; 
   sos_Class_type      ct;

   // create new sos_CompModif and append it to change_comps
   // it is not checked, whether the same component was deleted or
   // modified before. This connection will be made after the version
   // is complete and the adaption functions are created
   // but: err_raise, if component was modified before 
   ct   = sos_Class_type::make(self.get_modified_type()); 
   cpad = sos_CompModif::create(self.container());
   cpad.set_modified_component(comp);
   cpad.set_change(PROT_ADD);

   self.get_change_comps().append(cpad); 

   self.save_offsets(FALSE);

   self.set_change(PROT_MOD);

   TT(prot_H, T_LEAVE);
}   

// *************************************************************************
void sos_ClassModif::del_comp(sos_Comp_descr comp)
// *************************************************************************
{
   T_PROC("sos_ClassModif::del_comp");
   TT(prot_H, T_ENTER);

   sos_CompModif_List  compmodifs;
   sos_CompModif       cpmod;
   sos_CompModif       cpm;
   sos_Bool            found = FALSE;
   sos_Int             index = 0;
   sos_Int             cpm_pos;
  

   // check whether the component was added or modified during this version
   // if it was added, the sos_CompModif can be deleted from the list
   // if it was modified, change of that sos_CompModif is set to
   // PROT_DEL. By doing this, it is garantieed that the modification
   // executed before the deletion are not lost if the Komponet is
   // added again (undo)
   compmodifs = self.get_change_comps();
 
   agg_iterate(compmodifs, cpmod) 
   {
      index ++;

      if (cpmod.get_modified_component() == comp)
      {  // remember this CompModif 
         found   = TRUE;
         cpm     = cpmod;       
         cpm_pos = index;
      }
   } agg_iterate_end(compmodifs, cpmod);


   if (found)
   {  if (cpm.get_change() == PROT_ADD)
      {  // component was added during this version
         compmodifs.remove(cpm_pos);
         cpm.destroy();
      }
      else if (cpm.get_change() == PROT_MOD)
      {  // component was modified during this version
         cpm.set_change(PROT_DEL);
         self.save_offsets(TRUE);
      }
      else
      {  // component was deleted during this version
         err_raise(err_SYS, err_PROT_COMP_DEL, prot_ERR, FALSE);
      }
   }
   else
   {  // component was never touched during this version
      cpmod = sos_CompModif::create(self.container());
      cpmod.set_modified_component(comp);
      cpmod.set_change(PROT_DEL);

      compmodifs.append(cpmod);
 
      self.save_offsets(FALSE);
 
      self.set_change(PROT_MOD);
   }      
   TT(prot_H, T_LEAVE);
}


// *************************************************************************
void sos_ClassModif::mod_comp(sos_Comp_descr comp)
// *************************************************************************
{
   T_PROC("sos_ClassModif::mod_comp");
   TT(prot_H, T_ENTER);

   sos_Container       cnt;
   sos_CompModif_List  compmodifs;
   sos_CompModif       cpmod;
   sos_CompModif       cpm;
   sos_Bool            found = FALSE;


   // check whether the component was touched before 
   // if it was added, the modification has not to be logged. If the addition
   // was an undo of a previous deletion, the old values have already 
   // been saved at deletion time.  
   // if it was modified, nothing has to be done 
   // if it was deleted, error in MTA 
   compmodifs = self.get_change_comps();
 
   agg_iterate(compmodifs, cpmod)
   {
      if (cpmod.get_modified_component() == comp)
      {  // remember this CompModif
         found   = TRUE;
         cpm     = cpmod;
      }
   } agg_iterate_end(compmodifs, cpmod);
 
 
   if (found)
   {  if (cpm.get_change() != PROT_ADD &&
	  cpm.get_change() != PROT_MOD) 
      {  // component is deleted 
         err_raise(err_SYS, err_PROT_COMP_DEL, prot_ERR, FALSE);
      }
      else 
        self.save_offsets(FALSE);
   }
   else
   {  // the component has neither been added, deleted nor modified during
      // this version. So log the modification

      cnt   = self.container();
      cpmod = sos_CompModif::create(cnt);
      cpmod.set_modified_component(comp);
      cpmod.set_change(PROT_MOD);

      cpmod.set_old_name       (sos_String::clone(comp.get_name(), cnt));
      cpmod.set_old_type       (comp.get_get_method().get_result_type());
      cpmod.set_old_init       (comp.get_init_expr());
      cpmod.set_old_local      (comp.get_is_local());
      cpmod.set_old_static     (FALSE);  // <af>
      cpmod.set_old_value      (comp.get_is_value());
      cpmod.set_old_get_scope  (comp.get_get_method().get_kind());
      cpmod.set_old_set_scope  (comp.get_set_method().get_kind());
       
      compmodifs.append(cpmod);

      self.save_offsets(FALSE);           

      self.set_change(PROT_MOD);
   }
   TT(prot_H, T_LEAVE);
}
   
// *************************************************************************
void sos_ClassModif::add_meth(sos_Method meth)
// *************************************************************************
{
   T_PROC("sos_ClassModif::add_meth");
   TT(prot_H, T_ENTER);

   // it is assumed that the method was not modified before

   if (self.get_deleted_methods().is_element(meth))
      self.get_deleted_methods().remove(meth);
   else
   {  self.get_added_methods().insert(meth);
      self.set_change(PROT_MOD);
   }
   TT(prot_H, T_LEAVE);
}

// *************************************************************************
void sos_ClassModif::del_meth(sos_Method meth)
// *************************************************************************
{
   T_PROC("sos_ClassModif::del_meth");
   TT(prot_H, T_ENTER);

   if (self.get_added_methods().is_element(meth))
      self.get_added_methods().remove(meth);
   else
   {  // if the method was modified before, the modification log stays 
      // in modified_methods to allow an undo of the deletion
      // (not implemented yet)
      // but there is no connection established between deleted and 
      // added methods because it don't matter whether a method is
      // readded or not
      self.get_deleted_methods().insert(meth);

      self.set_change(PROT_MOD);
   }
   TT(prot_H, T_LEAVE);
}

// *************************************************************************
sos_MethodModif sos_ClassModif::mod_meth(sos_Method meth)
// *************************************************************************
{
   T_PROC("sos_ClassModif::mod_meth");
   TT(prot_H, T_ENTER);

   sos_Container        cnt = self.container();
   sos_MethodModif_List mm_list;
   sos_MethodModif      mm  = sos_MethodModif::make(NO_OBJECT);
   sos_MethodModif      act_mm;

   // method is new or deleted, nothing has to be done
   if (self.get_added_methods().is_element(meth) ||
       self.get_deleted_methods().is_element(meth))
   {  TT(prot_H, T_LEAVE);
      return sos_MethodModif::make(NO_OBJECT);
   }

   // search whether method was modified before, nothing has to be done
   mm_list = self.get_modified_methods();
   agg_iterate(mm_list, act_mm)
   {  if (act_mm.get_modified_method() == meth)  
         mm = act_mm;
   } agg_iterate_end(mm_list, act_mm);

   if (INVALID(mm))
   {
      // first modification of method in this version 
      mm = sos_MethodModif::create(cnt);
      mm.set_modified_method(meth);
      self.get_modified_methods().append(mm);

      mm.set_old_name        (sos_String::clone(meth.get_name(), cnt));
      mm.set_old_kind        (meth.get_kind());
      mm.set_old_result_type (meth.get_result_type());
      mm.set_old_abstract    (meth.get_is_abstract());
      mm.set_old_definite    (meth.get_is_definite());
      mm.set_old_static      (meth.get_is_static()); 
 
      self.set_change(PROT_MOD);
   } 

   TT(prot_H, T_LEAVE);
   return mm;
} 
    

// *************************************************************************
void sos_CompModif::local_finalize(sos_CompModif cpm) 
// *************************************************************************
{
   T_PROC("sos_CompModif::local_finalize");
   TT(prot_M, T_ENTER);

   if (VALID(cpm.get_old_name()))
      cpm.get_old_name().destroy();

   TT(prot_M, T_LEAVE);
} 
